﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace WordSearcher
{
    public partial class FormMain : Form
    {
        /// <summary>
        /// Флаг разрешения распознавания
        /// </summary>
        bool enableRecognize = false;
        /// <summary>
        /// Размер битмапа со словом "Указ"
        /// к этому размеру подгоняются все отсалдьные распознаваемые битмапы
        /// </summary>
        private static Size imSize = new Size(65, 25);
        /// <summary>
        /// Состояние формы
        /// </summary>
        private FormState formState = FormState.Empty;
        /// <summary>
        /// Масив битмапов сос словами текста
        /// </summary>
        private List<Bitmap> words;
        /// <summary>
        /// Объект класса для распознавания(персептрон)
        /// </summary>
        private Recognizer r = new Recognizer(imSize, 750, 2);

        /// <summary>
        /// Метод для делания активными неактивными кнопок управления
        /// в зависимости от этапа обработк изображения
        /// </summary>
        /// <param name="fs">текущее состояние</param>
        private void ButtonsEnabled(FormState fs)
        {
            formState = fs;
            switch (fs)
            {
                case FormState.Empty:
                    buttonOpen.Enabled = true;
                    buttonSegment.Enabled = false;
                    buttonRecognize.Enabled = false;
                    buttonTeach.Enabled = false;
                    buttonLoadTeaching.Enabled = false;
                    buttonSaveTeaching.Enabled = false;
                                        
                    pictureBoxMain.Image = null;
                    dataGridViewSegments.Rows.Clear();
                    break;
                case FormState.Open:
                    buttonOpen.Enabled = true;
                    buttonSegment.Enabled = true;
                    buttonRecognize.Enabled = false;
                    buttonTeach.Enabled = false;
                    buttonLoadTeaching.Enabled = false;
                    buttonSaveTeaching.Enabled = false;

                    dataGridViewSegments.Rows.Clear();
                    break;
                case FormState.Segmented:
                    buttonOpen.Enabled = true;
                    buttonSegment.Enabled = true;
                    if (enableRecognize)
                        buttonRecognize.Enabled = true;
                    else
                        buttonRecognize.Enabled = false;
                    buttonTeach.Enabled = true;
                    buttonLoadTeaching.Enabled = true;
                    buttonSaveTeaching.Enabled = false;
                    break;
                case FormState.Teached:
                    buttonOpen.Enabled = true;
                    buttonSegment.Enabled = false;
                    buttonRecognize.Enabled = true;
                    buttonTeach.Enabled = false;
                    buttonLoadTeaching.Enabled = false;
                    buttonSaveTeaching.Enabled = true;

                    enableRecognize = true;
                    break;
                case FormState.Deserialized:
                    buttonOpen.Enabled = true;
                    buttonSegment.Enabled = false;
                    buttonRecognize.Enabled = true;
                    buttonTeach.Enabled = false;
                    buttonLoadTeaching.Enabled = false;
                    buttonSaveTeaching.Enabled = true;

                    enableRecognize = true;
                    break;
                case FormState.Recognized:
                    buttonOpen.Enabled = true;
                    buttonSegment.Enabled = false;
                    buttonRecognize.Enabled = true;
                    buttonTeach.Enabled = false;
                    buttonLoadTeaching.Enabled = true;
                    buttonSaveTeaching.Enabled = true;
                    break;
            }
        }
        
        public FormMain()
        {
            InitializeComponent();
        }

        private void buttonOpen_Click(object sender, EventArgs e)
        {
            try
            {
                Bitmap b;
                if (openFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    //Если изобраение имеет индексированный формат 
                    //то переводим его в обычный, потомутчо с индексировнным не работат
                    //setpixel 
                    b = new Bitmap(openFileDialog1.FileName);
                    if (b.PixelFormat == System.Drawing.Imaging.PixelFormat.Format1bppIndexed ||
                        b.PixelFormat == System.Drawing.Imaging.PixelFormat.Format4bppIndexed ||
                        b.PixelFormat == System.Drawing.Imaging.PixelFormat.Format8bppIndexed)
                        b = new Bitmap(b);
                    pictureBoxMain.Image = b;
                    this.ButtonsEnabled(FormState.Open);

                }
                else
                {
                    this.ButtonsEnabled(FormState.Empty);
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        private void buttonSegment_Click(object sender, EventArgs e)
        {
            //Получаем набор битмапов соответствующих словам текста
            words = Segmentation.GetWords((Bitmap)pictureBoxMain.Image);
            dataGridViewSegments.RowCount = words.Count;
            int i  =0;
            //Перебираем слов и отображаем в таблице
            foreach (Bitmap word in words) 
            {
                dataGridViewSegments.Rows[i].Cells[0].Value = word;
                i++;
            }
            ButtonsEnabled(FormState.Segmented);
        }

        private void buttonRecognize_Click(object sender, EventArgs e)
        {
            int i = 0;
            foreach (Bitmap word in words) 
            {
                dataGridViewSegments.Rows[i].Cells[1].Value = r.Recognize(Recognizer.NormalizeBitmap(word,imSize));
                i++;
            }
            ButtonsEnabled(FormState.Recognized);
        }

        private void buttonTeach_Click(object sender, EventArgs e)
        {
            //Перебираем слова и обучаем ими
            //т.к. класса для распознаваня два то 
            //первый класс обучаем просто изображение слова
            //а второй обучаем противопорложным изображение(т.е инвертируем цвета исходного изображения)
            for (int i = 0; i < 5; i++)
            {
                foreach (Bitmap word in words)
                {
                    r.Teach(Recognizer.NormalizeBitmap(word, imSize), 0);
                    r.Teach(Recognizer.InverseBitmap(Recognizer.NormalizeBitmap(word, imSize)), 1);
                }
            }
            ButtonsEnabled(FormState.Teached);
        }

        private void buttonSaveTeaching_Click(object sender, EventArgs e)
        {
            r.SerializeParams();
            ButtonsEnabled(FormState.Serialized);
        }

        private void buttonLoadTeaching_Click(object sender, EventArgs e)
        {
            r.DeserializeParams();
            ButtonsEnabled(FormState.Deserialized);
        }

        private void FormMain_Load(object sender, EventArgs e)
        {
            this.ButtonsEnabled(FormState.Empty);
        }
    }

    // <summary>
    /// Состояния изображения
    /// </summary>
    enum FormState
    {
        /// <summary>
        /// изображение не открыто
        /// </summary>
        Empty,
        /// <summary>
        /// Изображение открыто
        /// </summary>
        Open,
        /// <summary>
        /// Сегментировано
        /// </summary>
        Segmented,
        /// <summary>
        /// Персептрон обучен
        /// </summary>
        Teached,
        /// <summary>
        /// Параметры персептрона сохранены
        /// </summary>
        Serialized,
        /// <summary>
        /// Параметры персептрона загружены
        /// </summary>
        Deserialized,
        /// <summary>
        /// Распознано
        /// </summary>
        Recognized
    }
}
