﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Windows.Forms;
using System.IO;
using System.Runtime.Serialization.Formatters.Binary;


namespace WordSearcher
{
    /// <summary>
    /// Реализует распозноание изображений
    /// на базе персептрона
    /// </summary>
    class Recognizer
    {
        /// <summary>
        /// матрица знаков входов персептрона 
        /// </summary>
        private int[,] xa;
        /// <summary>
        /// Массив лямд
        /// </summary>
        private int[,] l;

        /// <summary>
        /// Массив имен классов
        /// </summary>
        private string[] classes = {"Указ", 
                                   "Не указ"};

        /// <summary>
        /// Массив имен классов
        /// </summary>
        public string[] ClassesList 
        {
            get { return classes; }
        }

        /// <summary>
        /// Инициализирует xa-матрицу
        /// </summary>
        /// <param name="sz">размер изображения</param>
        /// <param name="aCount">количесвто а-элементов</param>
        /// <param name="lCount">количесвто классов</param>
        public Recognizer(Size sz, int aCount, int lCount) 
        {
            Random r = new Random();
            //Создание матрцы ха
            xa = new int[sz.Height * sz.Width, aCount];
            
            //Создание матрицы лямд
            l = new int[lCount,aCount];

            //Первоначальная
            //иницализация лямд еденицами
            for (int i = 0; i < l.GetLength(0); i++) 
            {
                for (int j = 0; j < l.GetLength(1); j++) 
                {
                    l[i, j] = 1;
                }
            }

            //заполнение матрицы
            //для каждого рецептора(строчки) 
            //назначаетя только один а-элемент(столбец) со знаком + или -
            for (int i = 0; i < xa.GetLength(0); i++)
            {
                xa[i, r.Next(aCount)] = (int)Math.Pow(-1, r.Next(1, 3));
            }
        }
        
        /// <summary>
        /// Обучение персептрона
        /// </summary>
        /// <param name="b">битмап для обучения</param>
        /// <param name="classindex">имя класса к ккоторому относиться изображение</param>
        public void Teach(Bitmap b, int classindex) 
        {
            int[] x = new int[b.Height * b.Width];

            int k = 0;
            //Инициализация входных рецепторов
            for (int i = 0; i < b.Width; i++) 
            {
                for (int j = 0; j < b.Height; j++) 
                {
                    if (b.GetPixel(i, j) == Color.FromArgb(0, 0, 0))
                        x[k] = 1;
                    k++;
                }
            }

            //Вектор сумм рецепторов
            int[] sumx = new int[xa.GetLength(1)];
            //Вектор выходов А-элементов
            int[] outa = new int[xa.GetLength(1)];


            //суммирование сигналов от рецепторов 
            for (int i = 0; i < xa.GetLength(1); i++) 
            {
                for (int j = 0; j < xa.GetLength(0); j++) 
                {
                    sumx[i] += x[j] * xa[j, i];
                }
                //Если сумма больше нуля выход а элемента 1
                if (sumx[i] > 0)
                    outa[i] = 1;
            }


            //изменение коэфициетов лямда
            for (int i = 0; i < outa.Length; i++) 
            {
                //Если а-элемент возбужден то изменяем лямды
                if (outa[i] == 1)
                {
                    //перебор всех классов
                    for (int j = 0; j < l.GetLength(0); j++)
                    {
                        //Увеличение на 1 лямд для класса который обучается
                        //и уменьшение для всех осатльных
                        if (classindex == j)
                            l[j, i]++;
                        else
                            l[j, i]--;
                    }
                }
            }    
        }

        /// <summary>
        /// Распознавание изобржения
        /// </summary>
        /// <param name="b">битмап изображения</param>
        /// <returns>имя класса к которому отнесено изображение</returns>
        public string Recognize(Bitmap b) 
        {
            int[] x = new int[b.Height * b.Width];

            int k = 0;
            //Инициализация входных рецепторов
            for (int i = 0; i < b.Width; i++)
            {
                for (int j = 0; j < b.Height; j++)
                {
                    if (b.GetPixel(i, j) == Color.FromArgb(0, 0, 0))
                        x[k] = 1;
                    k++;
                }
            }

            //Вектор суммрецепторов
            int[] sumx = new int[xa.GetLength(1)];
            //Вектор выходов А-элементов
            int[] outa = new int[xa.GetLength(1)];

            //суммирование сигналов от рецепторов 
            for (int i = 0; i < xa.GetLength(1); i++)
            {
                for (int j = 0; j < xa.GetLength(0); j++)
                {
                    sumx[i] += x[j] * xa[j, i];
                }
                //Если сумма больше нуля выход а элемента 1
                if (sumx[i] > 0)
                    outa[i] = 1;
            }

            //Создание масива значений сумматоров 
            //каждый для отдельного класса
            int[] sum = new int[l.GetLength(0)];
            //Нахождение значений сумматоров для каждого класса
            for (int i = 0; i < sum.Length; i++) 
            {
                for (int j = 0; j < xa.GetLength(1); j++) 
                {
                    sum[i] += outa[j] * l[i, j];
                }
            }

            //нахождение максимального значения сумматор
            //именно оно соответствует распознанному классу
            int max = sum[0];
            int maxindex = 0;
            for (int i = 1; i < sum.Length; i++) 
            {
                if (max < sum[i])
                {
                    max = sum[i];
                    maxindex = i;
                }
            }

            //Возвращается имя класса с максимальным значением сумматора
            return classes[maxindex];
        }

        /// <summary>
        /// Сериализация массива лямд(сохранение в файл) для сохранения обученяи персептрона
        /// </summary>
        public void SerializeParams() 
        {
            try
            {
                BinaryFormatter bf = new BinaryFormatter();
                FileStream fs = new FileStream("l.dat", FileMode.Create);
                bf.Serialize(fs, l);
                fs.Close();

                bf = new BinaryFormatter();
                fs = new FileStream("xa.dat", FileMode.Create);
                bf.Serialize(fs, xa);
                fs.Close();
            }
            catch (Exception ex) 
            {
                MessageBox.Show(ex.Message);
            }
        }
        
        /// <summary>
        /// Десериализация массива лямд(чтение из файла) 
        /// </summary>
        public void DeserializeParams()
        {
            try
            {
                BinaryFormatter bf = new BinaryFormatter();
                FileStream fs = new FileStream("l.dat", FileMode.Open);
                l = (int[,])bf.Deserialize(fs);
                fs.Close();

                bf = new BinaryFormatter();
                fs = new FileStream("xa.dat", FileMode.Open);
                xa = (int[,])bf.Deserialize(fs);
                fs.Close();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// Подгонка битмапа по размеру и его бинаризация
        /// </summary>
        /// <param name="b">входной битмап</param>
        /// <param name="sz">новый размер битмапа</param>
        /// <returns>нормализованный битмап</returns>
        public static Bitmap NormalizeBitmap(Bitmap b, Size sz) 
        {
            //Подгонка размера
            Bitmap inImg = new Bitmap(b, sz);
            //Создание выходного битмапа на основе подогнанного
            Bitmap outImg = new Bitmap(inImg);

            //находим среднее значение яркости
            int sum = 0;
            for (int i = 0; i < outImg.Width; i++) 
            {
                for (int j = 0; j < outImg.Height; j++) 
                {
                    Color cl = ((Bitmap)inImg).GetPixel(i,j);
                    sum += (cl.R + cl.G + cl.B) / 3;
                }
            }
            int sredn = sum / (inImg.Width * inImg.Height);
            //Просматриваем изображнеи и бинаризуем его
            for (int i = 0; i < outImg.Width; i++)
            {
                for (int j = 0; j < outImg.Height; j++)
                {
                    Color cl = ((Bitmap)inImg).GetPixel(i,j);
                    int gray = (cl.R + cl.G + cl.B) / 3;
                    if (gray > sredn)
                        outImg.SetPixel(i, j, Color.FromArgb(255, 255, 255));
                    else
                        outImg.SetPixel(i, j, Color.FromArgb(0, 0, 0));
                }
            }

            return outImg;
        }

        /// <summary>
        /// Инверсия цвета битмапа
        /// </summary>
        /// <param name="b"></param>
        /// <returns></returns>
        public static Bitmap InverseBitmap(Bitmap b) 
        {
            Bitmap outImg = new Bitmap(b.Width, b.Height);

            for (int i = 0; i < b.Width; i++) 
            {
                for (int j = 0; j < b.Height; j++) 
                {
                    Color c = b.GetPixel(i,j);
                    outImg.SetPixel(i, j, Color.FromArgb(255 - c.R, 255 - c.G, 255 - c.B));
                }
            }

            return outImg;
        }
    }

}
