﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;

namespace WordSearcher
{
    class Segmentation
    {
        /// <summary>
        /// Разбиение битмапа с текстоми на строки
        /// </summary>
        /// <param name="b">исходный битмап</param>
        /// <returns>коллекция строк</returns>
        public static List<Bitmap> GetStrings(Bitmap text) 
        {
            List<Bitmap> strs = new List<Bitmap>();
            List<int> whiteLineIndexes = new List<int>();
            
            //Находим все белые горзонатльные линии на ихображении
            //и запоминаем их индексы
            for (int j = 0; j < text.Height; j++) 
            {
                bool whiteLineFound = true;
                for (int i = 0; i < text.Width; i++) 
                {
                    if (text.GetPixel(i, j) != Color.FromArgb(255, 255, 255))
                    {
                        whiteLineFound = false;
                        break;
                    }
                }
                if (whiteLineFound)
                    whiteLineIndexes.Add(j);
            }

            //Выделение строк между белыми несоседними линиями
            for (int i = 0; i < whiteLineIndexes.Count-1; i++) 
            {
                if (whiteLineIndexes[i + 1] - whiteLineIndexes[i] > 4) 
                {
                    strs.Add(text.Clone(
                        new Rectangle(
                            0, 
                            whiteLineIndexes[i],
                            text.Width,
                            whiteLineIndexes[i + 1] - whiteLineIndexes[i]+1),
                        System.Drawing.Imaging.PixelFormat.Format24bppRgb));
                }
            }

            return strs;
        }

        /// <summary>
        /// Получить список слов отдельной строки
        /// </summary>
        /// <param name="str">битмап со строкой текста</param>
        /// <returns>спсиок слов строки</returns>
        public static List<Bitmap> GetStringWords(Bitmap str) 
        {
            List<Bitmap> words = new List<Bitmap>();
            List<int> whiteLineIndexes = new List<int>();

            //Находим все белые вертикальные линии на изображении
            //и запоминаем их индексы
            for (int i = 0; i < str.Width; i++)
            {
                bool whiteLineFound = true;
                for (int j = 0; j < str.Height; j++)
                {
                    if (str.GetPixel(i, j).R < 100)
                    {
                        whiteLineFound = false;
                        break;
                    }
                }
                if (whiteLineFound)
                    whiteLineIndexes.Add(i);
            }

            //Ширина пробела
            int spaceWidth = 0;
            int sum = 0;
            int n = 0;
            //Вычисление ширины пробела
            for (int i = 0; i < whiteLineIndexes.Count - 1; i++)
            {
                int d = whiteLineIndexes[i + 1] - whiteLineIndexes[i];
                if (d > 1)
                {
                    sum += d;
                    n++;
                }
            }
            //Ширина пробела необходимо при дальнейшем выделении слов
            //коэф. подобран вручную
            spaceWidth = (int)Math.Round(sum * 0.45 / n + 0.1);
            //начальная координата слова
            int wordBegin = 0;
            //конечная координат слова
            int wordEnd = 0;
            //флаг указывающий на то найденно ли начало слова или нет
            //перволдится обратно в фолс после нахождения конца слова
            bool wordFound = false;
            //Счетчик ширины белой полоски
            int whiteWidth = 0;
            //Выделение слов
            for (int i = 0; i < whiteLineIndexes.Count - 1; i++)
            {
                //если линии не соседние и флаг wordFound фолс т.е.
                //слово еще не найдено
                //запоминаем координату певрой линии это будет 
                //координатой началом слова
                if ((whiteLineIndexes[i + 1] - whiteLineIndexes[i] > 1) &&
                    !wordFound)
                {
                    //обнуление счетчика идущих подряд белыхз линий
                    whiteWidth = 0;
                    //флаг найденного слова в тру
                    wordFound = true;
                    //инициализируем начальную координату слова
                    wordBegin = whiteLineIndexes[i];
                }
                //инициализируем конечную координату слова
                //если найдены не сосдение линии
                //но не обрезаем битмап и не добавлям его в коллекцию
                //т.к. необходисмо зделать проверку на ширину пробела
                if ((whiteLineIndexes[i + 1] - whiteLineIndexes[i] > 1) &&
                                    wordFound)
                {
                    whiteWidth = 0;
                    wordEnd = whiteLineIndexes[i + 1];
                }
                //Если найденны соседние белые линии
                //инкремируем счетчик белых линий и сравниваем ширину идущих подрд белых линий
                //с ранее высчитаной средней шириной пробела
                if (whiteLineIndexes[i + 1] - whiteLineIndexes[i] == 1)
                {
                    whiteWidth++;
                    if ((whiteWidth >= spaceWidth) &&
                        (wordEnd - wordBegin > 1))
                    {                        
                        //Обрезаем и добавляем слово в коллекцию
                        words.Add(TrimBitmap(
                            str.Clone(
                        new Rectangle(
                            wordBegin,
                            0,
                            wordEnd - wordBegin + 1,
                            str.Height),
                        System.Drawing.Imaging.PixelFormat.Format24bppRgb)
                                             )
                                );
                        //обнуляем счетчики
                        //и  флаги
                        whiteWidth = 0;
                        wordFound = false;
                        wordBegin = 0;
                        wordEnd = 0;
                    }
                }
            }
 
            return words;
        }

        /// <summary>
        /// Получить битмапы всех слов в тексте
        /// </summary>
        /// <param name="text">битмап с текстом</param>
        /// <returns>коллекция всех слов в тексте</returns>
        public static List<Bitmap> GetWords(Bitmap text) 
        {
            List<Bitmap> strs = GetStrings(text);
            List<Bitmap> words = new List<Bitmap>();

            foreach (Bitmap str in strs) 
            {
                foreach (Bitmap word in GetStringWords(str)) 
                {
                    words.Add(word);
                }
            }
            return words;
        }
        
        /// <summary>
        /// Обрезка белых полей вокруг изображения на битмапе
        /// </summary>
        /// <param name="bmp"></param>
        /// <returns></returns>
        public static Bitmap TrimBitmap(Bitmap bmp)
        {
            int left = 0;
            int right = 0;
            int top = 0;
            int bottom = 0;

            bool go = true;
            //проход сверху
            for (int j = 0; (j < bmp.Height) && go; j++)
            {
                for (int i = 0; (i < bmp.Width) && go; i++)
                {
                    if (bmp.GetPixel(i, j) != Color.FromArgb(255, 255, 255))
                    {
                        go = false;
                        top = j;
                    }
                }
            }

            go = true;
            //проход снизу
            for (int j = bmp.Height - 1; (j >= 0) && go; j--)
            {
                for (int i = 0; (i < bmp.Width) && go; i++)
                {
                    if (bmp.GetPixel(i, j) != Color.FromArgb(255, 255, 255))
                    {
                        go = false;
                        bottom = j;
                    }
                }
            }

            go = true;
            //проход слева  
            for (int i = 0; (i < bmp.Width) && go; i++)
            {
                for (int j = 0; (j < bmp.Height) && go; j++)
                {
                    if (bmp.GetPixel(i, j) != Color.FromArgb(255, 255, 255))
                    {
                        go = false;
                        left = i;
                    }
                }
            }

            go = true;
            //проход спарва
            for (int i = bmp.Width - 1; (i >= 0) && go; i--)
            {
                for (int j = 0; (j < bmp.Height) && go; j++)
                {
                    if (bmp.GetPixel(i, j) != Color.FromArgb(255, 255, 255))
                    {
                        go = false;
                        right = i;
                    }
                }
            }

            return bmp.Clone(new Rectangle(left, top, right - left + 1, bottom - top + 1), System.Drawing.Imaging.PixelFormat.Format24bppRgb);
        }
    }
}
